# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''pytest tests for the BruteForce class in the search.py melody file'''

import pytest
from melody.search import BruteForce


def test_run(capsys):
    '''check that bruteforce runs when we provide a valid set of inputs'''
    from melody.inputs import Choice
    inputs = [Choice(name="c",
                     inputs=["c1", "c2"])]

    def function(input_var):
        '''test function'''
        return 0, input_var
    search_method = BruteForce(function=function, inputs=inputs)
    search_method.run()
    out, _ = capsys.readouterr()
    assert "[{'c': 'c1'}] 0 [{'c': 'c1'}]" in out
    assert "[{'c': 'c2'}] 0 [{'c': 'c2'}]" in out


@pytest.mark.xfail(reason=("bug : state support is work in progress"))
def test_state_run(capsys, monkeypatch):
    '''check that bruteforce runs when we provide a valid set of inputs
    with state'''
    from melody.inputs import Choice
    choice = Choice(name="c", inputs=["c1", "c2"])
    monkeypatch.setattr(choice, "_state", True)
    inputs = [choice]

    def function(input_var):
        '''test function'''
        return 0, input_var
    state = "state"
    search_method = BruteForce(function=function, inputs=inputs, state=state)
    search_method.run()
    out, _ = capsys.readouterr()
    assert "[{'c': 'c1'}] 0 [{'c': 'c1'}]" in out
    assert "[{'c': 'c2'}] 0 [{'c': 'c2'}]" in out


def test_func_too_few_ret_args():
    '''check that an exception is raised if too few function return
    arguments are provided'''
    from melody.inputs import Choice
    choice = Choice(name="c", inputs=["c1", "c2"])
    inputs = [choice]
    def function(inputs):
        '''test function'''
        return inputs
    search_method = BruteForce(inputs, function)
    with pytest.raises(RuntimeError) as excinfo:
        search_method.run()
    assert "function must return 2 values" in str(excinfo.value)


def test_func_too_many_ret_args():
    '''check that an exception is raised if too many function return
    arguments are provided'''
    from melody.inputs import Choice
    choice = Choice(name="c", inputs=["c1", "c2"])
    inputs = [choice]
    def function(inputs):
        '''test function'''
        return inputs, 0, 0
    search_method = BruteForce(inputs=inputs, function=function)
    with pytest.raises(RuntimeError) as excinfo:
        search_method.run()
    assert "function must return 2 values" in str(excinfo.value)



# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''pytest tests for the Name class in the inputs.py melody file'''

from melody.inputs import Choice


def test_choice_class_value_vanilla():
    '''check that the Choice class stores the required name and inputs
    arguments correctly'''
    test_values = ["test1", "test2", "test3"]
    test_name = "err"
    choice = Choice(name=test_name, inputs=test_values)
    assert len(choice.options) == len(test_values)
    for idx, _ in enumerate(test_values):
        assert choice.options[idx] == test_values[idx]
    assert choice.name == test_name


def test_choice_class_value_pre():
    '''check that the Choice class returns the specified inputs modified
    by the pre value if it is provided'''
    test_name = "umm"
    test_values = ["test1", "test2", "test3"]
    pre_value = "pre_"
    choice = Choice(test_name, test_values, pre=pre_value)
    assert len(choice.options) == len(test_values)
    for idx, _ in enumerate(test_values):
        assert choice.options[idx] == pre_value+test_values[idx]



# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''melody is a simple generic tuning library primarily designed for
the optimisation of software on hpc architectures'''


class Melody(object):
    ''' The main class '''

    def __init__(self, function=None, method=None, state=None, inputs=None):
        self._function = function
        self._method = method
        self._state = state
        self._inputs = inputs

    @property
    def function(self):
        '''Return the function associated with this instance of melody'''
        return self._function

    @function.setter
    def function(self, function):
        '''Set the function associated with this instance of melody'''
        self._function = function

    @property
    def method(self):
        '''Return the optimisation/search method associated with this instance
        of melody'''
        return self._method

    @method.setter
    def method(self, method):
        '''Set the optimisation/search method associated with this instance of
melody'''
        self._method = method

    @property
    def inputs(self):
        '''Return the input search parameters for the associated function'''
        return self._inputs

    @inputs.setter
    def inputs(self, inputs):
        '''Set the input search parameters for the associated function'''
        self._inputs = inputs

    def search(self):
        '''Start the optimisation/search using the supplied optimisation
        method with the supplied inputs for the supplied function'''
        search = self._method(inputs=self._inputs, function=self._function,
                              state=self._state)
        search.run()



# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''pytest tests for the create_input function in the inputs.py melody file'''

import os
import pytest
from melody.inputs import create_input

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
TEMPLATE_DIR = os.path.join(SCRIPT_DIR, "template")
TEMPLATE_NAME = "template.txt"
EXPECTED = "result"
OPTION = [{"KEY": EXPECTED}]


def test_create_input():
    ''' test that we can modify a template file correctly '''
    result = create_input(OPTION, TEMPLATE_NAME,
                          template_location=TEMPLATE_DIR)
    assert EXPECTED in result


def test_option_wrong_format():
    '''check that we raise an exception if the input option is not in the
    expected format (a dictionary within a list).'''
    option = ["fred"]
    with pytest.raises(RuntimeError) as excinfo:
        _ = create_input(option, TEMPLATE_NAME,
                         template_location=TEMPLATE_DIR)
    assert "Expecting a dictionary" in str(excinfo.value)


def test_non_existant_template():
    '''test that we raise an exception if the specified template does not
    exist'''
    template_name = "invalid.txt"
    with pytest.raises(RuntimeError) as excinfo:
        _ = create_input(OPTION, template_name,
                         template_location=TEMPLATE_DIR)
    assert "template '"+template_name+"' not found" in str(excinfo.value)


def test_missing_template_dir():
    '''test that we raise an exception if the specified template location
    does not exist'''
    template_dir = os.path.join(SCRIPT_DIR, "invalid")
    with pytest.raises(RuntimeError) as excinfo:
        _ = create_input(OPTION, TEMPLATE_NAME,
                         template_location=template_dir)
    assert "template '"+TEMPLATE_NAME+"' not found" in str(excinfo.value)



# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''pytest tests for the SearchMethod class in the search.py melody file'''

import pytest
from melody.search import SearchMethod


def test_vanilla():
    '''check that initial required values are stored appropriately if they
    are provided when the class is initialised and that optional
    values are set to None'''
    from melody.inputs import Fixed
    inputs = [Fixed(name="dummy", value="dummy")]

    def function(arg):
        '''test function'''
        return 0, arg
    search_method = SearchMethod(inputs=inputs, function=function)
    assert search_method.inputs is inputs
    assert search_method.function is function
    assert search_method.state is None


def test_init_state():
    '''check that state is stored appropriately if it is
    provided when the class is initialised'''
    from melody.inputs import Fixed
    inputs = [Fixed(name="dummy", value="dummy")]

    def function(arg):
        '''test function'''
        return 0, arg
    state = "envy"
    search_method = SearchMethod(inputs=inputs, function=function, state=state)
    assert search_method.inputs is inputs
    assert search_method.function is function
    assert search_method.state is state


def test_set_state():
    '''check that state is stored appropriately if it is set '''
    from melody.inputs import Fixed
    inputs = [Fixed(name="dummy", value="dummy")]

    def function(arg):
        '''test function'''
        return 0, arg
    state = "envy"
    search_method = SearchMethod(inputs, function)
    search_method.state = state
    assert search_method.inputs is inputs
    assert search_method.function is function
    assert search_method.state is state


def test_run_method():
    '''check that a notimplemented exception is raised if the run method
    is called'''
    from melody.inputs import Fixed
    inputs = [Fixed(name="dummy", value="dummy")]

    def function(arg):
        '''test function'''
        return 0, arg
    search_method = SearchMethod(inputs, function)
    with pytest.raises(NotImplementedError) as excinfo:
        search_method.run()
    assert "Run method should be implemented" in str(excinfo.value)


def test_inputs_not_iterable():
    '''check that an exception is raised inputs is not iteratable'''
    inputs = 7

    def function(arg):
        '''test function'''
        return 0, arg
    with pytest.raises(RuntimeError) as excinfo:
        _ = SearchMethod(inputs=inputs, function=function)
    assert "inputs should be iterable" in str(excinfo.value)


def test_inputs_base_class():
    '''check that an exception is raised if any of the inputs are not
    subclasses of the inputs class'''
    inputs = ["hello"]

    def function(arg):
        '''test function'''
        return 0, arg
    with pytest.raises(RuntimeError) as excinfo:
        _ = SearchMethod(inputs, function)
    assert "input should be a subclass of the Input class" \
        in str(excinfo.value)


def test_invalid_function():
    '''check that an exception is raised if an invalid function is provided'''
    from melody.inputs import Fixed
    inputs = [Fixed(name="dummy", value="dummy")]
    function = "invalid"
    with pytest.raises(RuntimeError) as excinfo:
        _ = SearchMethod(inputs=inputs, function=function)
    assert "provided function '"+function+"' is not callable" \
        in str(excinfo.value)


def test_invalid_function2():
    '''check that an exception is raised if too few arguments are provided
    to the function'''
    from melody.inputs import Fixed
    inputs = [Fixed(name="dummy", value="dummy")]

    def function():
        '''test function'''
        return 0, 0
    with pytest.raises(RuntimeError) as excinfo:
        _ = SearchMethod(inputs=inputs, function=function)
    assert "provided function should have one argument" in str(excinfo.value)


def test_invalid_function3():
    '''check that an exception is raised if too many arguments are provided
    to the function'''
    from melody.inputs import Fixed
    inputs = [Fixed(name="dummy", value="dummy")]

    def function(arg1, arg2):
        '''test function'''
        return arg1, arg2
    with pytest.raises(RuntimeError) as excinfo:
        _ = SearchMethod(inputs=inputs, function=function)
    assert "provided function should have one argument" in str(excinfo.value)



# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''This file contains a set of input classes that users can use to
define the space they want to search. These classes can also be
specialised for particular scenarios.'''


class Input(object):
    '''This is the base class for all types of input class.'''

    def __init__(self, name, options, state=None):
        self._state = state
        self._name = name
        self._options = options

    @property
    def state(self):
        '''Returns the state variable. This is 'None' by default.'''
        return self._state

    @property
    def name(self):
        '''Returns the name of the input class. All classes are expected to
        define a name'''
        return self._name

    @property
    def options(self):
        '''Returns the particular set of options that are valid for the
        instance of the input'''
        return self._options


class Fixed(Input):
    '''This class simply returns a single fixed value. It therefore does
    not add to the search space. Rather it is a convenience class.'''

    def __init__(self, name, value):
        options = [value]
        Input.__init__(self, name, options)


class Switch(Input):
    '''This class allows the user to specify two options where only one is
    valid at a time. It could be considered to be a special case of
    Choice, but it is so common that a separate class makes sense.'''

    def __init__(self, name, off=None, on=None):
        if not off and not on:
            raise RuntimeError(
                "A value must be provided for at least one of 'off' or 'on'")
        options = []
        if off:
            options.append(off)
        else:
            options.append("")
        if on:
            options.append(on)
        else:
            options.append("")
        Input.__init__(self, name, options)


class Choice(Input):
    '''This class allows the user to specify an unlimited number of
    options, where only one option is valid at a time.

    '''

    def __init__(self, name, inputs, pre=None):
        options = []
        for value in inputs:
            if pre:
                options.append(pre + value)
            else:
                options.append(value)
        Input.__init__(self, name, options)


class IntRange(Input):
    '''This class implements the integer version of the Range class,
    allowing a set of integer inputs to be defined by a low, high and
    step.'''

    def __init__(self, name, low, high, step):
        options = [i for i in range(low, high, step)]
        Input.__init__(self, name, options)


class FloatRange(Input):
    '''This class implements a range of floating point values,
    which are set by low, high and step values.'''

    def __init__(self, name, low, high, step):
        from numpy import arange
        options = [f for f in arange(low, high, step)]
        Input.__init__(self, name, options)


class Subsets(Input):
    '''This class allows all combinations of a particular set of values to
    be chosen, including, no values, individual values, pairs of values,
    triplets of values etc.'''

    def __init__(self, name, inputs):
        self._inputs = inputs
        self._options = []
        for k in range(len(inputs) + 1):
            self._recurse(inputs, [], depth=0, max_depth=k)
        Input.__init__(self, name, self._options)

    def _recurse(self, inputs, output, depth, max_depth):
        '''We work out all combinations using this internal recursion method'''
        if depth < max_depth:
            for index, option in enumerate(inputs):
                my_output = list(output)
                my_output.append(option)
                self._recurse(inputs[index + 1:], my_output, depth + 1,
                              max_depth)
        else:
            self._options.append(output)


def create_input(option, template_name, template_location="template"):

    '''create an input file using jinja2 by filling a template
    with the values from the option variable passed in.'''

    # restructure option list into jinja2 input format
    jinja2_input = {}
    for item in option:
        try:
            jinja2_input.update(item)
        except ValueError:
            raise RuntimeError(
                ("inputs.py, create_input : format of item '{0}' is not "
                 "supported. Expecting a dictionary.".format(str(item))))

    # load the template and fill it with the option variable contents
    import jinja2
    try:
        template_loader = jinja2.FileSystemLoader(searchpath=template_location)
        template_env = jinja2.Environment(loader=template_loader)
        template = template_env.get_template(template_name)
        output_text = template.render(jinja2_input)
    except jinja2.TemplateNotFound:
        raise RuntimeError("template '{0}' not found".format(template_name))
    # return the particular input file as a string
    return output_text



# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
"""Setup script. Used by easy_install and pip."""

from setuptools import setup, find_packages

PACKAGES = find_packages(where="src")

NAME = 'melody'
AUTHOR = 'Rupert Ford'
AUTHOR_EMAIL = 'rupert.ford@stfc.ac.uk'
URL = 'https://github.com/rupertford/melody'
DOWNLOAD_URL = 'https://github.com/rupertford/melody'
DESCRIPTION = 'lightweight python parameter search tool'
LONG_DESCRIPTION = open('README.md').read()
LICENSE = 'OSI Approved :: BSD 3-Clause License'
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Natural Language :: English',
    'Programming Language :: Python :: 2.7',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development',
    'Topic :: Utilities',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS']

MAJOR = 0
MINOR = 1
MICRO = 1
ISRELEASED = not True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

if __name__ == '__main__':

    setup(
        name=NAME,
        version=VERSION,
        author=AUTHOR,
        author_email=(AUTHOR_EMAIL),
        license=LICENSE,
        url=URL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        classifiers=CLASSIFIERS,
        packages=PACKAGES,
        package_dir={"": "src"},
        install_requires=['jinja2'])




# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''pytest tests for the Melody class in the main.py melody file'''

from melody.main import Melody


def test_melody_vanilla():
    '''check that initial values are set to None if they are not
    provided'''
    melody = Melody()
    assert melody.function is None
    assert melody.inputs is None
    assert melody.method is None
    assert melody._state is None


def test_melody_initial_values():
    '''check that values are stored correctly if provided at initialisation'''
    function = "function"
    method = "method"
    state = "state"
    inputs = "inputs"
    melody = Melody(function=function, method=method, state=state,
                    inputs=inputs)
    assert melody.function == function
    assert melody.inputs == inputs
    assert melody.method == method
    assert melody._state == state


def test_melody_set_values():
    '''check that values are stored correctly if directly set'''
    function = "function"
    method = "method"
    inputs = "inputs"
    melody = Melody()
    melody.function = function
    melody.method = method
    melody.inputs = inputs
    assert melody.function == function
    assert melody.inputs == inputs
    assert melody.method == method


def test_melody_search(capsys):
    '''check that melody search runs when we provide a valid set of inputs'''
    from melody.inputs import Choice
    from melody.search import BruteForce
    inputs = [Choice(name="c",
                     inputs=["c1", "c2"])]

    def function(input_var):
        '''test function'''
        return 0, input_var
    search_method = BruteForce
    melody = Melody(method=search_method, inputs=inputs, function=function)
    melody.search()
    out, _ = capsys.readouterr()
    assert "[{'c': 'c1'}] 0 [{'c': 'c1'}]" in out
    assert "[{'c': 'c2'}] 0 [{'c': 'c2'}]" in out



# BSD 3-Clause License
#
# Copyright (c) 2017, Science and Technology Facilities Council
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Author R. Ford STFC Daresbury Lab.
#
'''This module provides the internally supported Melody search
methods. A search method is expected to conform to the SearchMethod
base class'''


class SearchMethod(object):
    '''A utility baseclass for different search/optimisation methods'''

    def __init__(self, inputs, function, state=None):
        self._inputs = inputs
        self._check_inputs()
        self._function = function
        self._check_function()
        self._state = state

    def _check_inputs(self):
        ''' make some basic checks on the inputs to make sure they are valid'''
        try:
            _ = self._inputs[0]
        except TypeError:
            raise RuntimeError(
                "inputs should be iterable but found type='{0}', value="
                "'{1}'".format(type(self._inputs), str(self._inputs)))
        from melody.inputs import Input
        for check_input in self._inputs:
            if not isinstance(check_input, Input):
                raise RuntimeError(
                    "input should be a subclass of the Input class but "
                    "found type='{0}', value='{1}'".format(type(check_input),
                                                           str(check_input)))

    def _check_function(self):
        ''' make some basic checks on the function to make sure it is valid'''
        # note, callable is valid for Python 2 and Python 3.2 onwards but
        # not inbetween
        if not callable(self._function):
            raise RuntimeError(
                "provided function '{0}' is not callable".
                format(str(self._function)))
        from inspect import getargspec
        arg_info = getargspec(self._function)
        if len(arg_info.args) != 1:
            print str(arg_info)
            raise RuntimeError(
                "provided function should have one argument but found "
                "{0}".format(len(arg_info.args)))
        # I don't think I can statically determine how many arguments
        # a function will return so will have to check after calling
        # the function

    def run(self):
        '''Ensure that any subclass implements this method'''
        raise NotImplementedError("Run method should be implemented")

    @property
    def function(self):
        '''Return the function associated with this instance of search
        method'''
        return self._function

    @property
    def inputs(self):
        '''Return the input search parameters for this instance of search
        method'''
        return self._inputs

    @property
    def state(self):
        '''Return the state for this instance of search method'''
        return self._state

    @state.setter
    def state(self, state):
        '''Set the state for this instance of search method'''
        self._state = state


class BruteForce(SearchMethod):
    '''A search method that tests all input options'''

    def __init__(self, inputs, function, state=None):
        SearchMethod.__init__(self, inputs, function, state=state)

    def run(self):
        ''' perform the search over inputs'''
        self._recurse(self.inputs, [])

    def _recurse(self, inputs, output):
        '''internal recursion routine called by the run method that generates
        all input combinations'''
        if inputs:
            my_input = inputs[0]
            name = my_input.name
            if my_input.state:
                my_options = my_input.options(self.state)
            else:
                my_options = my_input.options
            for option in my_options:
                my_output = list(output)
                my_output.append({name: option})
                self._recurse(inputs[1:], my_output)
        else:
            try:
                valid, result = self._function(output)
            except ValueError:
                raise RuntimeError("function must return 2 values")
            print output, valid, result
